/*
 * blu-save helps you to backup your bluray discs
 * Copyright 2019 Yunxiang Li
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "bluray.h"
#include "file/file.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

int64_t total = 0;

void dump_file(BLURAY *bd, const char *rel_path, const char *cache_path) {
    /* open input file */
    BD_FILE_H *fp_in = bd_open_file_dec(bd, rel_path);
    if (!fp_in) {
        fprintf(stderr, "failed to open %s\n", rel_path);
        return;
    }

    /* open output file */
    FILE *fp_out = fopen(cache_path, "wb");
    if (!fp_out) {
        fprintf(stderr, "failed to create cache file %s\n", cache_path);
        file_close(fp_in);
        return;
    }

    /* write to output file */
    const size_t buf_size = 6144 * 60;
    uint8_t buf[buf_size];
    int64_t wrote = 0;
    int64_t got;
    do {
        got = file_read(fp_in, buf, buf_size);

        /* we'll call write(fp, buf, 0) after EOF. It is used to check for
         * errors.
         */
        if (got < 0 || fwrite(buf, 1, got, fp_out) != got) {
            fprintf(stderr, "failed to cache file %s\n", rel_path);
            fclose(fp_out);
            file_close(fp_in);
            remove(cache_path);
            return;
        }
        wrote += got;
        fprintf(stdout, "%ld MiB.\r", (wrote + total) / 1024 / 1024);
    } while (got > 0);

    total += wrote;
    fprintf(stdout, "cached %s, total %ld MiB.\n", cache_path,
            total / 1024 / 1024);

    fclose(fp_out);
    file_close(fp_in);
}

void dump_dir(BLURAY *bd, const char *base_dir, const char *out_dir) {
    BD_DIRENT ep;
    BD_DIR_H *dp = (BD_DIR_H *)bd_open_dir(bd, base_dir);

    if (dp) {
        if (!dir_read(dp, &ep)) { // non-empty dir
            char *base = malloc(strlen(base_dir) + !base_dir[0] + 255 + 1);
            char *out = malloc(strlen(out_dir) + !out_dir[0] + 255 + 1);

            mkdir(out_dir, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH); // drwxrwxr-x
            do {
                char *name = ep.d_name;
                if (base_dir[0]) {
                    strcpy(base, base_dir);
                    strcat(base, DIR_SEP);
                    strcat(base, name);
                } else {
                    strcpy(base, name);
                }
                if (out_dir[0]) {
                    strcpy(out, out_dir);
                    strcat(out, DIR_SEP);
                    strcat(out, name);
                } else {
                    strcpy(out, name);
                }
                if (strstr(name, ".")) {
                    dump_file(bd, base, out);
                } else {
                    dump_dir(bd, base, out);
                }
            } while (!dir_read(dp, &ep));
            free(out);
            free(base);
        }
        dir_close(dp);
    } else {
        fprintf(stderr, "failed to open %s\n", base_dir);
    }
}

int main(int argc, char *argv[]) {
    BLURAY *bd;
    char *disc_dir, *out_dir, *key;

    if (argc == 3) {
        disc_dir = argv[1];
        out_dir = argv[2];
        key = NULL;
    } else if (argc == 4) {
        disc_dir = argv[1];
        out_dir = argv[2];
        key = argv[3];
    } else {
        fprintf(stderr, "Usage:\n   %s <disc_path> <out_dir> [keyfile]\n",
                argv[0]);
        return -1;
    }

    /* open disc */
    bd = bd_open(disc_dir, key);
    if (!bd) {
        fprintf(stderr, "failed to open %s\n", argv[1]);
        return -1;
    }

    /* dump */
    dump_dir(bd, "", out_dir);

    /* close disc */
    bd_close(bd);

    return 0;
}
